#ifndef DUNE_SUBGRIDENTITY_HH
#define DUNE_SUBGRIDENTITY_HH

/** \file
* \brief The SubGridEntity class
*/

#include <dune/grid/common/referenceelements.hh>


namespace Dune {


// Forward declarations

template<int codim, int dim, class GridImp>
class SubGridEntity;

template<int codim, class GridImp>
class SubGridEntityPointer;

template<int codim, PartitionIteratorType pitype, class GridImp>
class SubGridLevelIterator;

template<class GridImp>
class SubGridLevelIntersectionIterator;

template<class GridImp>
class SubGridLeafIntersectionIterator;

template<class GridImp>
class SubGridHierarchicIterator;




template<int codim, int dim, class GridImp>
class SubGridMakeableEntity :
    public GridImp::template Codim<codim>::Entity
{
    public:
    
        // The codimension of this entitypointer wrt the host grid
        enum {CodimInHostGrid = GridImp::HostGridType::dimension - GridImp::dimension + codim};
    
        // EntityPointer to the equivalent entity in the host grid
        typedef typename GridImp::HostGridType::Traits::template Codim<CodimInHostGrid>::EntityPointer HostGridEntityPointer;
        
        
        //! \todo Please doc me !
        SubGridMakeableEntity(const GridImp* subGrid, const HostGridEntityPointer& hostEntity) :
            GridImp::template Codim<codim>::Entity (SubGridEntity<codim, dim, const GridImp>(subGrid,hostEntity)),
            subGrid_(subGrid)
        {}
        
        
        //! \todo Please doc me !
        void setToTarget(const HostGridEntityPointer& hostEntity) {
            this->realEntity.setToTarget(hostEntity);
        }
        
        
        //! \todo Please doc me !
        const HostGridEntityPointer& getTarget() {
            return this->realEntity.hostEntity_;
        }
    
        
    private:
        
        const GridImp* subGrid_;
};


//**********************************************************************
//
// --SubGridEntity
// --Entity
//
/** \brief The implementation of entities in a SubGrid
*   \ingroup SubGrid
*
*  A Grid is a container of grid entities. An entity is parametrized by the codimension.
*  An entity of codimension c in dimension d is a d-c dimensional object.
*
*/
template<int codim, int dim, class GridImp>
class SubGridEntity :
    public EntityDefaultImplementation <codim,dim,GridImp,SubGridEntity>
{
    friend class SubGridMakeableEntity<codim,dim,GridImp>;

    template <class GridImp_>
    friend class SubGridLevelIndexSet;

    template <class GridImp_>
    friend class SubGridLeafIndexSet;

    template <class GridImp_>
    friend class SubGridLocalIdSet;

    template <class GridImp_>
    friend class SubGridGlobalIdSet;

    template <class GridImp_, int EntityDim>
    friend class IndexSetter;

    template <int codim_, class GridImp_>
    friend class SubGridEntityPointer;

    private:
        
        typedef typename GridImp::ctype ctype;
        
        // The codimension of this entitypointer wrt the host grid
        enum {CodimInHostGrid = GridImp::HostGridType::dimension - GridImp::dimension + codim};

        // EntityPointer to the equivalent entity in the host grid
        typedef typename GridImp::HostGridType::Traits::template Codim<CodimInHostGrid>::EntityPointer HostGridEntityPointer;
        

    public:
    
        typedef typename GridImp::template Codim<codim>::Geometry Geometry;
    
        
        //! Constructor for an entity in a given grid level
        SubGridEntity(const GridImp* subGrid, const HostGridEntityPointer& hostEntity) :
            hostEntity_(hostEntity),
            subGrid_(subGrid),
            geo_(0),
            geoInFather_(0)
        {}
        
    
        //! \todo Please doc me !
        SubGridEntity(const SubGridEntity& original) :
            hostEntity_(original.hostEntity_),
            subGrid_(original.subGrid_),
            geo_(0),
            geoInFather_(0)
        {}
    
        
        //! Destructor
        ~SubGridEntity()
        {
            if (geo_!=0)
            {
                delete geo_;
                geo_ = 0;
            }
            if (geoInFather_!=0)
            {
                delete geoInFather_;
                geoInFather_ = 0;
            }
        }
        
        
        //! \todo Please doc me !
        SubGridEntity& operator=(const SubGridEntity& original)
        {
            if (this != &original)
            {
                if (geo_!=0)
                {
                    delete geo_;
                    geo_ = 0;
                }
                if (geoInFather_!=0)
                {
                    delete geoInFather_;
                    geoInFather_ = 0;
                }
                subGrid_ = original.subGrid_;
                hostEntity_ = original.hostEntity_;
            }
            return *this;
        }
    
    
        //! level of this element
        int level () const {
            return hostEntity_->level();
        }
    
        
        /** \brief The partition type for parallel computing
        */
        PartitionType partitionType () const {
            return hostEntity_->partitionType();
        }
    
        
        /** Intra-element access to entities of codimension cc > codim. Return number of entities
        * with codimension cc.
        */
        template<int cc> int count () const{
            return hostEntity_->template count<cc>();
        }
        
        
        //! geometry of this entity
        const Geometry& geometry () const
        {
            if (geo_==0)
                geo_ = new SubGridMakeableGeometry<dim-codim,GridImp::dimensionworld,GridImp>(hostEntity_->geometry());
            return *geo_;
        }
    
        GeometryType type () const
        {
            return hostEntity_->type();
        }
        
        HostGridEntityPointer hostEntity_;
    
        
    private:
    
        //! \todo Please doc me !
        void setToTarget(const HostGridEntityPointer& target)
        {
            if(geo_!=0)
            {
                delete geo_;
                geo_ = 0;
            }
            if (geoInFather_!=0)
            {
                delete geoInFather_;
                geoInFather_ = 0;
            }
            hostEntity_ = target;
        }
    
        
        const GridImp* subGrid_;
        
        //! the current geometry
        mutable SubGridMakeableGeometry<dim-codim,GridImp::dimensionworld,GridImp> *geo_;
        mutable SubGridMakeableGeometry<dim-codim,dim,GridImp> *geoInFather_;
};




//***********************
//
//  --SubGridEntity
//
//***********************
/** \brief Specialization for codim-0-entities.
* \ingroup SubGrid
*
* This class embodies the topological parts of elements of the grid.
* It has an extended interface compared to the general entity class.
* For example, Entities of codimension 0  allow to visit all neighbors.
*/
template<int dim, class GridImp>
class SubGridEntity<0,dim,GridImp> :
    public EntityDefaultImplementation<0,dim,GridImp, SubGridEntity>
{
    public:
    
        // The codimension of this entitypointer wrt the host grid
        enum {CodimInHostGrid = GridImp::HostGridType::dimension - GridImp::dimension};
    
        // EntityPointer to the equivalent entity in the host grid
        typedef typename GridImp::HostGridType::Traits::template Codim<CodimInHostGrid>::EntityPointer HostGridEntityPointer;
        
        typedef typename GridImp::template Codim<0>::Geometry Geometry;
    
        typedef typename GridImp::template Codim<0>::LocalGeometry LocalGeometry;
    
        //! The Iterator over intersections on this level
        typedef SubGridLevelIntersectionIterator<GridImp> LevelIntersectionIterator;
    
        //! The Iterator over intersections on the leaf level
        typedef SubGridLeafIntersectionIterator<GridImp> LeafIntersectionIterator;
    
        //! Iterator over descendants of the entity
        typedef SubGridHierarchicIterator<GridImp> HierarchicIterator;
        
        
        //! Constructor for an entity in a given grid level
        SubGridEntity(const GridImp* subGrid, const HostGridEntityPointer& hostEntity) :
            subGrid_(subGrid),
            geo_(0),
            geoInFather_(0),
            hostEntity_(hostEntity)
        {}
        
        
        //! \todo Please doc me !
        SubGridEntity(const SubGridEntity& original) :
            subGrid_(original.subGrid_),
            geo_(0),
            geoInFather_(0),
            hostEntity_(original.hostEntity_)
        {}
    
        
        //! Destructor
        ~SubGridEntity()
        {
            if (geo_!=0)
            {
                delete geo_;
                geo_ = 0;
            }
            if (geoInFather_!=0)
            {
                delete geoInFather_;
                geoInFather_ = 0;
            }
        }
        
        
        //! \todo Please doc me !
        SubGridEntity& operator=(const SubGridEntity& original)
        {
            if (this != &original)
            {
                if (geo_!=0)
                {
                    delete geo_;
                    geo_ = 0;
                }
                if (geoInFather_!=0)
                {
                    delete geoInFather_;
                    geoInFather_ = 0;
                }
                subGrid_ = original.subGrid_;
                hostEntity_ = original.hostEntity_;
            }
            return *this;
        }
    
        
        //! Level of this element
        int level () const
        {
            return hostEntity_->level();
        }
    
        
        /** \brief The partition type for parallel computing */
        PartitionType partitionType () const {
            return hostEntity_->partitionType();
        }
    
        
        //! Geometry of this entity
        const Geometry& geometry () const
        {
            if (geo_==0)
            {
                geo_ = new SubGridMakeableGeometry<dim,GridImp::dimensionworld,GridImp>(hostEntity_->geometry());
            }
            return *geo_;
        }
    
        
        /** \brief Return the number of subEntities of codimension cc.
        */
        template<int cc>
        int count () const
        {
            return hostEntity_->template count<cc>();
        }
        
        
        /** \brief Provide access to sub entity i of given codimension. Entities
        *  are numbered 0 ... count<cc>()-1
        */
        template<int cc>
        typename GridImp::template Codim<cc>::EntityPointer entity (int i) const{
            return SubGridEntityPointer<cc,GridImp>(subGrid_, hostEntity_->template entity<cc>(i));
        }
    
        
        //! First level intersection
        SubGridLevelIntersectionIterator<GridImp> ilevelbegin () const{
            return SubGridLevelIntersectionIterator<GridImp>(subGrid_,
                                                             hostEntity_->ilevelbegin());
        }
    
        
        //! Reference to one past the last neighbor
        SubGridLevelIntersectionIterator<GridImp> ilevelend () const{
            return SubGridLevelIntersectionIterator<GridImp>(subGrid_,
                                                             hostEntity_->ilevelend());
        }
    
        
        //! First leaf intersection
        SubGridLeafIntersectionIterator<GridImp> ileafbegin () const{
            if (!isLeaf()) return ileafend();
            return SubGridLeafIntersectionIterator<GridImp>(subGrid_,
                hostEntity_->ilevelbegin(), hostEntity_->ilevelend());
        }
    
        
        //! Reference to one past the last leaf intersection
        SubGridLeafIntersectionIterator<GridImp> ileafend () const{
            return SubGridLeafIntersectionIterator<GridImp>(subGrid_,
                hostEntity_->ilevelend(), hostEntity_->ilevelend());
        }
    
        
        //! returns true if Entity has NO children
        bool isLeaf() const {
            return subGrid_->template isLeaf<0>(*hostEntity_);
        }
    
        
        //! Inter-level access to father element on coarser grid.
        //! Assumes that meshes are nested.
        SubGridEntityPointer<0,GridImp> father () const {
            return SubGridEntityPointer<0,GridImp>(subGrid_, hostEntity_->father());
        }
    
    
        /** \brief Location of this element relative to the reference element element of the father.
        * This is sufficient to interpolate all dofs in conforming case.
        * Nonconforming may require access to neighbors of father and
        * computations with local coordinates.
        * On the fly case is somewhat inefficient since dofs  are visited several times.
        * If we store interpolation matrices, this is tolerable. We assume that on-the-fly
        * implementation of numerical algorithms is only done for simple discretizations.
        * Assumes that meshes are nested.
        */
        const LocalGeometry& geometryInFather () const {
            if (geoInFather_==0)
                geoInFather_ = new SubGridMakeableGeometry<dim,dim,GridImp>(hostEntity_->geometryInFather());
            return *geoInFather_;
        }
    
        
        /** \brief Inter-level access to son elements on higher levels<=maxlevel.
        * This is provided for sparsely stored nested unstructured meshes.
        * Returns iterator to first son.
        */
        SubGridHierarchicIterator<GridImp> hbegin (int maxLevel) const
        {
            return SubGridHierarchicIterator<const GridImp>(subGrid_, *this, maxLevel);
        }
    
        
        //! Returns iterator to one past the last son
        SubGridHierarchicIterator<GridImp> hend (int maxLevel) const
        {
            return SubGridHierarchicIterator<const GridImp>(subGrid_, *this, maxLevel, true);
        }
        
        
        //! \todo Please doc me !
        bool wasRefined () const
        {
            return subGrid_->wasRefined(*this);
        }
        
        
        //! \todo Please doc me !
        bool mightBeCoarsened () const
        {
            return true;
        }
    
        
        // /////////////////////////////////////////
        //   Internal stuff
        // /////////////////////////////////////////
    
        
        //! \todo Please doc me !
        void setToTarget(const HostGridEntityPointer& target)
        {
            if(geo_!=0)
            {
                delete geo_;
                geo_ = 0;
            }
            if (geoInFather_!=0)
            {
                delete geoInFather_;
                geoInFather_ = 0;
            }
            hostEntity_ = target;
        }
        
        
        const GridImp* subGrid_;
        
        //! the current geometry
        mutable SubGridMakeableGeometry<dim,GridImp::dimensionworld,GridImp> *geo_;
        
        //! \todo Please doc me !
        mutable SubGridMakeableGeometry<dim,dim,GridImp> *geoInFather_;

        //! \todo Please doc me !
        HostGridEntityPointer hostEntity_;
    
        
    private:
    
        typedef typename GridImp::ctype ctype;

}; // end of SubGridEntity codim = 0


} // namespace Dune


#endif
