#ifndef DUNE_SUBGRID_INDEXSETS_HH
#define DUNE_SUBGRID_INDEXSETS_HH

/** \file
* \brief The index and id sets for the SubGrid class
*/

#include <vector>

namespace Dune {


template <class GridImp>
struct SubGridLevelIndexSetTypes
{
    //! The types
    template<int cd>
    struct Codim
    {
        template<PartitionIteratorType pitype>
        struct Partition
        {
        typedef typename GridImp::Traits::template Codim<cd>::template Partition<pitype>::LevelIterator Iterator;
        };
    };
};




/** \todo Take the index types from the host grid */
template<class GridImp>
class SubGridLevelIndexSet :
    public IndexSetDefaultImplementation<GridImp,SubGridLevelIndexSet<GridImp>,SubGridLevelIndexSetTypes<GridImp> >
{
    public:
        
        typedef typename remove_const<GridImp>::type::HostGridType HostGrid;
        typedef typename HostGrid::Traits::GlobalIdSet HostIdSet;
        typedef typename HostIdSet::IdType HostIdType;
    
        enum {dim = GridImp::dimension};
        
        typedef IndexSetDefaultImplementation<GridImp,SubGridLevelIndexSet<GridImp>,SubGridLevelIndexSetTypes<GridImp> > Base;
        
        /** \brief Default constructor
        * Unfortunately we can't force the user to init grid_ and level_, because
        * UGGridLevelIndexSets are meant to be stored in an array.
        *
        */
        SubGridLevelIndexSet ()
        {}
        
        
        //! get index of an entity
        template<int codim>
        int index (const typename GridImp::Traits::template Codim<codim>::Entity& e) const
        {
            return grid_->indexStorage.template levelIndex<codim>(e, level_);
        }
        
        
        //! get index of subEntity of a codim 0 entity
        template<int codim>
        int subIndex (const typename GridImp::Traits::template Codim<0>::Entity& e, int i) const
        {
            return grid_->indexStorage.template levelSubIndex<codim>(e, i, level_);
        }
        
        
        //! get number of entities of given codim, type and on this level
        int size (int codim) const {
            return grid_->indexStorage.levelSize(codim, level_);
        }
        
        
        //! get number of entities of given codim, type and on this level
        int size (GeometryType type) const
        {
            return grid_->indexStorage.levelSize(type, level_);
        }
        
        
        /** \brief Deliver all geometry types used in this grid */
        const std::vector<GeometryType>& geomTypes (int codim) const
        {
            return grid_->indexStorage.levelGeomTypes(codim, level_);
        }

    /** @brief Return true if the given entity is contained in \f$E\f$.
     */
    template<class EntityType>
    bool contains (const EntityType& e) const
    {
        return grid_ == grid_->getRealImplementation(e).subGrid_ 
            && level_ == grid_->getRealImplementation(e).hostEntity_->level();
    }
        
        
        //! one past the end on this level
        template<int cd, PartitionIteratorType pitype>
        typename Base::template Codim<cd>::template Partition<pitype>::Iterator begin () const
        {
            return grid_->template lbegin<cd,pitype>(level_);
        }
        
        
        //! Iterator to one past the last entity of given codim on level for partition type
        template<int cd, PartitionIteratorType pitype>
        typename Base::template Codim<cd>::template Partition<pitype>::Iterator end () const
        {
            return grid_->template lend<cd,pitype>(level_);
        }
        
        
        /** \todo Currently we support only vertex and element indices */
        void update(const GridImp& grid, int level)
        {
            grid_ = &grid;
            level_ = level;
        }
        
        
        GridImp* grid_;
        
        int level_;
};




template <class GridImp>
struct SubGridLeafIndexSetTypes
{
    //! The types
    template<int cd>
    struct Codim
    {
        template<PartitionIteratorType pitype>
        struct Partition
        {
        typedef typename GridImp::Traits::template Codim<cd>::template Partition<pitype>::LeafIterator Iterator;
        };
    };
};




template<class GridImp>
class SubGridLeafIndexSet :
    public IndexSetDefaultImplementation<GridImp,SubGridLeafIndexSet<GridImp>,SubGridLeafIndexSetTypes<GridImp> >
{
    public:
    
        typedef typename remove_const<GridImp>::type::HostGridType HostGrid;
        typedef typename HostGrid::Traits::GlobalIdSet HostIdSet;
        typedef typename HostIdSet::IdType HostIdType;
    
        /*
        * We use the remove_const to extract the Type from the mutable class,
        * because the const class is not instantiated yet.
        */
        enum {dim = remove_const<GridImp>::type::dimension};
        
        typedef IndexSetDefaultImplementation<GridImp,SubGridLeafIndexSet<GridImp>,SubGridLeafIndexSetTypes<GridImp> > Base;
        
        
        //! constructor stores reference to a grid and level
        SubGridLeafIndexSet ()
        {}
        
        
        //! get index of an entity
        /*
            We use the remove_const to extract the Type from the mutable class,
            because the const class is not instantiated yet.
        */
        template<int codim>
        int index (const typename remove_const<GridImp>::type::Traits::template Codim<codim>::Entity& e) const
        {
            return grid_->indexStorage.template leafIndex<codim>(e);
        }
        
        
        //! get index of subEntity of a codim 0 entity
        /*
            We use the remove_const to extract the Type from the mutable class,
            because the const class is not instantiated yet.
        */
        template<int codim>
        int subIndex (const typename remove_const<GridImp>::type::Traits::template Codim<0>::Entity& e, int i) const
        {
            return grid_->indexStorage.template leafSubIndex<codim>(e, i);
        }
        
        
        //! get number of entities of given type
        int size (GeometryType type) const
        {
            return grid_->indexStorage.leafSize(type);
        }
        
        
        //! get number of entities of given codim
        int size (int codim) const
        {
            return grid_->indexStorage.leafSize(codim);
        }
        
        
        /** \brief Deliver all geometry types used in this grid */
        const std::vector<GeometryType>& geomTypes (int codim) const
        {
            return grid_->indexStorage.leafGeomTypes(codim);
        }
        
        /** @brief Return true if the given entity is contained in \f$E\f$.
         */
        template<class EntityType>
        bool contains (const EntityType& e) const
        {
            enum { codim = EntityType::codimension };
            return (grid_->indexStorage.template leafIndex<codim>(e)>=0);
        }
        
        //! one past the end on this level
        template<int codim, PartitionIteratorType pitype>
        typename Base::template Codim<codim>::template Partition<pitype>::Iterator begin () const
        {
            return grid_->template leafbegin<codim,pitype>();
        }
        
        
        //! Iterator to one past the last entity of given codim on level for partition type
        template<int codim, PartitionIteratorType pitype>
        typename Base::template Codim<codim>::template Partition<pitype>::Iterator end () const
        {
            return grid_->template leafend<codim,pitype>();
        }
        
        
        /** \todo Currently we support only vertex and element indices */
        void update(const GridImp& grid)
        {
            grid_ = &grid;
        }
        
        
        GridImp* grid_;
};




template <class GridImp>
class SubGridGlobalIdSet :
    public IdSet<GridImp,SubGridGlobalIdSet<GridImp>,
        typename remove_const<GridImp>::type::HostGridType::Traits::GlobalIdSet::IdType>
{
    private:
    
        typedef typename remove_const<GridImp>::type::HostGridType HostGrid;
    
            
    public:
        //! constructor stores reference to a grid
        SubGridGlobalIdSet (const GridImp& g) : grid_(&g) {}
        
        //! define the type used for persistent indices
        typedef typename HostGrid::Traits::GlobalIdSet::IdType GlobalIdType;
        
        
        //! get id of an entity
        /*
        We use the remove_const to extract the Type from the mutable class,
        because the const class is not instantiated yet.
        */
        template<int cd>
        GlobalIdType id (const typename remove_const<GridImp>::type::Traits::template Codim<cd>::Entity& e) const
        {
            // Return id of the host entity
            return grid_->getHostGrid().globalIdSet().id(*grid_->getRealImplementation(e).hostEntity_);
        }
    
        
        //! get id of subEntity
        /*
            We use the remove_const to extract the Type from the mutable class,
            because the const class is not instantiated yet.
        */
        template<int cc>
        GlobalIdType subId (const typename remove_const<GridImp>::type::Traits::template Codim<0>::Entity& e, int i) const
        {
                // Return sub id of the host entity
                return grid_->getHostGrid().globalIdSet().template subId<cc>(*grid_->getRealImplementation(e).hostEntity_,i);
        }

        
        /** \todo Should be private */
        void update() {}

        
        const GridImp* grid_;
};




template<class GridImp>
class SubGridLocalIdSet :
    public IdSet<GridImp,SubGridLocalIdSet<GridImp>,
        typename remove_const<GridImp>::type::HostGridType::Traits::LocalIdSet::IdType>
{
    private:
    
        typedef typename remove_const<GridImp>::type::HostGridType HostGrid;

        
    public:
        //! define the type used for persistent local ids
        typedef typename HostGrid::Traits::LocalIdSet::IdType LocalIdType;

        
        //! constructor stores reference to a grid
        SubGridLocalIdSet (const GridImp& g) : grid_(&g) {}
    
        
        //! get id of an entity
        /*
            We use the remove_const to extract the Type from the mutable class,
            because the const class is not instantiated yet.
        */
        template<int cd>
        LocalIdType id (const typename remove_const<GridImp>::type::Traits::template Codim<cd>::Entity& e) const
        {
            // Return id of the host entity
            return grid_->getHostGrid().localIdSet().id(*grid_->getRealImplementation(e).hostEntity_);
        }
        
        
        //! get id of subEntity
        /*
        * We use the remove_const to extract the Type from the mutable class,
        * because the const class is not instantiated yet.
        */
        template<int cc>
        LocalIdType subId (const typename remove_const<GridImp>::type::Traits::template Codim<0>::Entity& e, int i) const
        {
            // Return sub id of the host entity
            return grid_->getHostGrid().localIdSet().template subId<cc>(*grid_->getRealImplementation(e).hostEntity_,i);
        }
        
        
        /** \todo Should be private */
        void update() {}

        
        const GridImp* grid_;
};


}  // namespace Dune


#endif
