#ifndef DUNE_SUBGRIDINTERSECTIONITERATOR_HH
#define DUNE_SUBGRIDINTERSECTIONITERATOR_HH

#include <list>
#include <tr1/memory>

/** \file
 * \brief The SubGridLeafIntersectionIterator and SubGridLevelIntersectionIterator classes
 */

namespace Dune {

template<class GridImp>
class SubGridIntersection
{
    enum {dim=GridImp::dimension};
    enum {dimworld=GridImp::dimensionworld};
    
    typedef typename GridImp::HostGridType::template Codim<0>::EntityPointer HostEntityPointer;
    typedef typename GridImp::HostGridType::template Codim<0>::Entity::LevelIntersectionIterator HostLevelIntersectionIterator;
    typedef typename GridImp::HostGridType::template Codim<1>::Geometry HostGeometry;
    typedef SubGridEntityPointer<0,GridImp> SubEntityPointer;
    typedef SubGridMakeableGeometry<dim-1,dimworld,GridImp> SubGeometry;
public:
    typedef typename GridImp::template Codim<0>::EntityPointer EntityPointer;
    typedef typename GridImp::template Codim<1>::LocalGeometry LocalGeometry;
    typedef typename GridImp::template Codim<1>::Geometry Geometry;
    
    SubGridIntersection(const GridImp* g,
        const HostEntityPointer & a, const HostEntityPointer & b,
        const HostLevelIntersectionIterator & hostIterator) :
        inside_(SubEntityPointer(g,a)), outside_(SubEntityPointer(g,b)), boundaryId_(-1),
        hostIterator_(hostIterator),
        intersectionGlobal_(NULL)
    {}

    SubGridIntersection(const GridImp* g,
        const HostEntityPointer & a, int b,
        const HostLevelIntersectionIterator & hostIterator) :
        inside_(SubEntityPointer(g,a)), outside_(SubEntityPointer(g,a)), boundaryId_(b),
        hostIterator_(hostIterator),
        intersectionGlobal_(NULL)
    {}

    ~SubGridIntersection() {
        if (intersectionGlobal_ != NULL) {
            delete intersectionGlobal_;
            intersectionGlobal_ = NULL;
        }
    }
    
    EntityPointer inside() const {
        return inside_;
    }
    
    EntityPointer outside() const {
        return outside_;
    }
    
    //! return true if across the edge a neighbor exists
    bool neighbor () const {
        return inside_ != outside_;
    }
        
    //! return true if intersection is with boundary.
    bool boundary () const {
        return boundaryId_ != -1;
    }
    
    //! return information about the boundary
    int boundaryId () const {
        return boundaryId_;
    }

    GeometryType type () const {
        DUNE_THROW(NotImplemented, "SubGridIntersection::type()");
    }

    //! intersection of codimension 1 of this neighbor with element where iteration started.
    //! Here returned element is in GLOBAL coordinates of the element where iteration started.
    const  Geometry& intersectionGlobal () const {
//        if (intersectionGlobal_ == NULL)
            intersectionGlobal_ = new SubGridMakeableGeometry<dim-1,dimworld,GridImp>(hostIterator_->intersectionGlobal());
        return *intersectionGlobal_;
    }

    //! intersection of codimension 1 of this neighbor with element where
    //! iteration started.
    //! Here returned element is in LOCAL coordinates of the element
    //! where iteration started.
    const  LocalGeometry& intersectionSelfLocal () const {
        DUNE_THROW(NotImplemented, "SubGridIntersection::intersectionSelfLocal()");
    }
        
    //! intersection of codimension 1 of this neighbor with element where iteration started.
    //! Here returned element is in LOCAL coordinates of neighbor
    const  LocalGeometry& intersectionNeighborLocal () const {
        DUNE_THROW(NotImplemented, "SubGridIntersection::intersectionNeighborLocal()");
    }
        
    bool operator== (const SubGridIntersection<GridImp> & other) const {
        return inside_ == other.inside_ && outside_ == other.outside_ && boundaryId_ == other.boundaryId_;
    }
    
private:
    EntityPointer inside_;
    EntityPointer outside_;
    int boundaryId_;
    const HostLevelIntersectionIterator hostIterator_;
    mutable SubGeometry* intersectionGlobal_;
};

template<class GridImp>
class SubGridBasicIntersectionIterator
{
private:
    
    enum {dim=GridImp::dimension};
    
    enum {dimworld=GridImp::dimensionworld};
    
    // The type used to store coordinates
    typedef typename GridImp::ctype ctype;
    
    typedef typename GridImp::HostGridType::template Codim<0>::Entity::LevelIntersectionIterator HostLevelIntersectionIterator;

public:
    
    typedef typename GridImp::template Codim<0>::EntityPointer EntityPointer;
    typedef typename GridImp::template Codim<1>::Geometry Geometry;
    typedef typename GridImp::template Codim<1>::LocalGeometry LocalGeometry;
    typedef typename GridImp::template Codim<0>::Entity Entity;
    typedef typename GridImp::HostGridType::Traits::LocalIdSet::IdType IdType;
    
    SubGridBasicIntersectionIterator(const GridImp* subGrid) :
        subGrid_(subGrid) {};

protected:

    enum fatherFaceType {
        coarsestFace, firstInternalFace
    };

    /** \brief Find the father face of currentFace, if it exists.
     */
    HostLevelIntersectionIterator
    getFatherFace(const HostLevelIntersectionIterator& currentFace,
        fatherFaceType fft) const {

        // get inside element
        typename GridImp::HostGridType::template Codim<0>::EntityPointer
            currentElement = currentFace->inside();

        // currentFace is the end if the inside element is on level 0
        if (currentElement->level() == 0 ||
            (fft == firstInternalFace &&
                currentFace->neighbor() &&
                subGrid_->template contains<0>(currentFace->outside())))
            return currentFace;

        // Get father intersection
        typename GridImp::HostGridType::template Codim<0>::EntityPointer
            fatherElement  = currentElement->father();

        // loop over all intersections of the father element and look for the father intersection
        HostLevelIntersectionIterator fatherFace = fatherElement->ilevelbegin();
        for (; fatherFace != fatherElement->ilevelend(); ++fatherFace)
            if (isSubFace(currentFace,fatherFace))
                return getFatherFace(fatherFace, fft);

        // we didn't find a father face, so this is end
        return currentFace;
    }

    /** \brief Return true if faceA is a father face of faceB or vice versa.

    The Dune grid interface does not support hierarchical iteration over faces.
    Therefore this implementation is hacky.  Precisely, this method returns true
    if the two faces are coplanar and share a common vertex id.
    */
    bool
    isSubFace(const HostLevelIntersectionIterator& faceA,
        const HostLevelIntersectionIterator& faceB) const
    {
        const typename GridImp::HostGridType::Traits::LocalIdSet &
            idSet_ = subGrid_->getHostGrid().localIdSet();

        // The following check assumes two things which may not always be justified
        //  - the intersections are planar
        //  - the intersection and its father intersection are coplanar.
        if (faceA->unitOuterNormal(FieldVector<ctype,dim-1>(0)) 
            * faceB->unitOuterNormal(FieldVector<ctype,dim-1>(0)) < 0.95)
            return false;
            
        // Check if face and father face share at least one common vertex.
        // There may be very exotic refinement rules where this is not true
        // Also, this assumes that intersections are faces...
        const ReferenceElement<ctype,dim>& refElementA
            = ReferenceElements<ctype, dim>::general(faceA->inside()->type());
        
        const ReferenceElement<ctype,dim>& refElementB
            = ReferenceElements<ctype, dim>::general(faceB->inside()->type());
        
        for (int i=0; i<refElementB.size(faceB->numberInSelf(),1,dim); i++) {
            
            int idB = refElementB.subEntity(faceB->numberInSelf(),1,i,dim);
            const IdType& VIdB = idSet_.template subId<dim>(*faceB->inside(),idB);
            
            for (int j=0; j<refElementA.size(faceA->numberInSelf(),1,dim); j++) {
                
                int idA = refElementA.subEntity(faceA->numberInSelf(),1,j,dim);
                
                // we have found a father face
                if (VIdB == idSet_.template subId<dim>(*faceA->inside(),idA))
                    return true;
            }
        }

        // No father face found
        return false;
    }
    
    const GridImp* subGrid_;
};
    
/** \brief Iterator over all element neighbors
 * \ingroup SubGrid
 * Mesh entities of codimension 0 ("elements") allow to visit all neighbors, where
 * a neighbor is an entity of codimension 0 which has a common entity of codimension 1
 * These neighbors are accessed via a IntersectionIterator. This allows the implement
 * non-matching meshes. The number of neighbors may be different from the number
 * of an element!
 */
template<class GridImp>
class SubGridLeafIntersectionIterator :
        public IntersectionIteratorDefaultImplementation <GridImp,SubGridLeafIntersectionIterator>,
        private SubGridBasicIntersectionIterator<GridImp>
{
private:
    
    enum {dim=GridImp::dimension};
    
    enum {dimworld=GridImp::dimensionworld};
    
    // The type used to store coordinates
    typedef typename GridImp::ctype ctype;
    
    typedef typename GridImp::HostGridType::template Codim<0>::LevelIntersectionIterator HostLevelIntersectionIterator;
    typedef typename GridImp::HostGridType::template Codim<1>::Geometry HostFaceGeometry;

    typedef SubGridIntersection<GridImp> Intersection;
	typedef std::list<Intersection> IntersectionList;
	typedef std::tr1::shared_ptr<IntersectionList> IntersectionListPtr;
    typedef SubGridEntityPointer<0,GridImp> SubEntityPointer;
    typedef typename GridImp::HostGridType::template Codim<0>::EntityPointer HostEntityPointer;
    typedef typename GridImp::HostGridType::template Codim<0>::HierarchicIterator HostHierarchicIterator;

public:
    
    typedef typename GridImp::template Codim<0>::EntityPointer EntityPointer;
    typedef typename GridImp::template Codim<1>::Geometry Geometry;
    typedef typename GridImp::template Codim<1>::LocalGeometry LocalGeometry;
    typedef typename GridImp::template Codim<0>::Entity Entity;

    SubGridLeafIntersectionIterator(const GridImp* subGrid,
        const HostLevelIntersectionIterator& hostIterator,
        const HostLevelIntersectionIterator& endIterator)
        : SubGridBasicIntersectionIterator<GridImp>(subGrid),
          hostIterator_(hostIterator), endIterator_(endIterator)
    {
        createSubIntersections();
    }
        
    //! The Destructor
    ~SubGridLeafIntersectionIterator() {};
        
    //! equality
    bool equals(const SubGridLeafIntersectionIterator<GridImp>& other) const {
        return (hostIterator_ == other.hostIterator_ &&
            ( ( subIntersections_->size() == 0 && other.subIntersections_->size() == 0) ||
                ( subIntersections_->size() != 0 && other.subIntersections_->size() != 0 &&
                    (*subIntersection_) == (*(other.subIntersection_)) ) ) );
    }
    
    //! prefix increment
    void increment() {
        ++subIntersection_;
        if(subIntersection_ == subIntersections_->end()) {
            ++hostIterator_;
            createSubIntersections();
        }
    }


    /** \brief Create a subgrid intersections that correspond to the current host grid intersection
     */
    void createSubIntersections()
    {
        subIntersections_ = IntersectionListPtr(new IntersectionList);
        if (hostIterator_ != endIterator_)
        {
            HostEntityPointer inside = hostIterator_->inside();

            if (hostIterator_->neighbor())
            {
                if (this->subGrid_->template contains<0>(hostIterator_->outside()))
                {
                    // bottom to top
                    pushChildFaces(inside, flipFace(hostIterator_));
                }
                else
                {
                    // top to bottom
                    HostLevelIntersectionIterator fatherFace = getFatherFace(hostIterator_, SubGridBasicIntersectionIterator<GridImp>::firstInternalFace);
                    HostEntityPointer father = fatherFace->outside();
                    
                    if (this->subGrid_->template contains<0>(father))
                    {
                        // neighbor
                        subIntersections_->push_back(Intersection(this->subGrid_, inside, father, hostIterator_));
                    }
                    else
                    {
                        // push_back boundary_intersection
                        int bid = 12345;
                        subIntersections_->push_back(Intersection(this->subGrid_, inside, bid, hostIterator_));
                    }
                }
            }
            else
            {
                if (hostIterator_->boundary())
                {
                    // push_back boundary_intersection
                    int bid = hostIterator_->boundaryId();
                    subIntersections_->push_back(Intersection(this->subGrid_, inside, bid, hostIterator_));
                }
                else
                {
                    // top to bottom
                    HostLevelIntersectionIterator fatherFace = getFatherFace(hostIterator_, SubGridBasicIntersectionIterator<GridImp>::firstInternalFace);
                    HostEntityPointer father = fatherFace->outside();
                    
                    if (this->subGrid_->template contains<0>(father))
                    {
                        // neighbor
                        subIntersections_->push_back(Intersection(this->subGrid_, inside, father, hostIterator_));
                    }
                    else
                    {
                        // push_back boundary_intersection
                        int bid = 12345;
                        subIntersections_->push_back(Intersection(this->subGrid_, inside, bid, hostIterator_));
                    }
                }
            }

        }

        subIntersection_ = subIntersections_->begin();
    }

    /** \brief Get Intersection corresponding to the one given as the argument, 
        but with inside and outside swapped
     */
    HostLevelIntersectionIterator
    flipFace(const HostLevelIntersectionIterator& currentFace) const
    {
        HostLevelIntersectionIterator i = currentFace->outside()->ilevelbegin();
        while (i->boundary() || (i->neighbor() && i->outside() != currentFace->inside()))
            ++i;
        return i;
    }
    
    void pushChildFaces(const HostEntityPointer& inside,
        const HostLevelIntersectionIterator currentFace)
    {
        // paranoia
        assert(!currentFace->boundary());

        // get inside element
        HostEntityPointer currentElement = currentFace->inside();

        // store currentFace if currentElement is a subgrid leaf entity
        if (this->subGrid_->template isLeaf<0>(*currentElement))
        {
            subIntersections_->push_back(Intersection(this->subGrid_, inside, currentElement, currentFace));
            return;
        }
        // currentFace is a boundary intersection if currentElement is not subgrid entity
        if (! this->subGrid_->template contains<0>(currentElement))
        {
            int bid = 12345;
            subIntersections_->push_back(Intersection(this->subGrid_, inside, bid, currentFace));
            return;
        }

        // currentFace is the end if the inside element is a leaf entity
        if (currentElement->isLeaf())
        {
            // we should have an intersection before reaching this point!
            assert(false);
            return;
        }

        // loop over all child Entities
        HostHierarchicIterator child = currentElement->hbegin(currentElement->level()+1);
        HostHierarchicIterator endchild = currentElement->hend(currentElement->level()+1);
        bool foundChild = false;
        for (; child != endchild; ++child)
        {

            // loop over all intersections of the child element and look for the nested intersection
            HostLevelIntersectionIterator childFace = child->ilevelbegin();
            for (; childFace != child->ilevelend(); ++childFace) {
                if (isSubFace(currentFace, childFace))
                {
                    pushChildFaces(inside, childFace);
                    foundChild = true;
                }
            }
        }

        if (! foundChild)
            DUNE_THROW(GridError, "Could not find nested intersection.");
    }
        
    //! return EntityPointer to the Entity on the inside of this intersection
    //! (that is the Entity where we started this Iterator)
    EntityPointer inside() const {
        return subIntersection_->inside();
    }
        
    //! return EntityPointer to the Entity on the outside of this intersection
    //! (that is the neighboring Entity)
    EntityPointer outside() const {
        return subIntersection_->outside();
    }
    
    //! return true if intersection is with boundary. \todo connection with
    //! boundary information, processor/outer boundary
    bool boundary () const {
        return subIntersection_->boundary();
    }
    
    //! return true if across the edge an neighbor on this level exists
    bool neighbor () const {
        return subIntersection_->neighbor();
    }
        
    //! return information about the Boundary
    int boundaryId () const {
        return subIntersection_->boundaryId();
    }

    /** \brief Get the geometry type */
    GeometryType type () const {
        return subIntersection_->type();
    }
            
    /** \brief Return true if this intersection is conforming */
    bool conforming () const {
        DUNE_THROW(NotImplemented, "SubGridLeafIntersectionIterator::conforming()");
        return false;
    }
              
    //! intersection of codimension 1 of this neighbor with element where
    //! iteration started.
    //! Here returned element is in LOCAL coordinates of the element
    //! where iteration started.
    const  LocalGeometry& intersectionSelfLocal () const {
        return subIntersection_->intersectionSelfLocal();
    }
        
    //! intersection of codimension 1 of this neighbor with element where iteration started.
    //! Here returned element is in GLOBAL coordinates of the element where iteration started.
    const  Geometry& intersectionGlobal () const {
        return subIntersection_->intersectionGlobal();
    }
        
    //! intersection of codimension 1 of this neighbor with element where iteration started.
    //! Here returned element is in LOCAL coordinates of neighbor
    const  LocalGeometry& intersectionNeighborLocal () const {
        return subIntersection_->intersectionNeighborLocal();
    }
        
    //! local number of codim 1 entity in self where intersection is contained in
    int numberInSelf () const {
        return hostIterator_->numberInSelf();
    }
        
    //! local number of codim 1 entity in neighbor where intersection is contained
    int numberInNeighbor () const {
        DUNE_THROW(NotImplemented, "SubGridLeafIntersectionIterator::numberInNeighbor");
    }
    
    //! return outer normal, this should be dependent on local
    //! coordinates for higher order boundary
    const FieldVector<ctype, dimworld> outerNormal (const FieldVector<ctype, dim-1>& local) const {
        return hostIterator_->outerNormal(local);
    }

    //! return outer normal, this should be dependent on local
    //! coordinates for higher order boundary
    const FieldVector<ctype, dimworld> integrationOuterNormal (const FieldVector<ctype, dim-1>& local) const {
        return  hostIterator_->integrationOuterNormal(local);
    }

    //! return outer normal, this should be dependent on local
    //! coordinates for higher order boundary
    const FieldVector<ctype, dimworld> unitOuterNormal (const FieldVector<ctype, dim-1>& local) const {
        return hostIterator_->unitOuterNormal(local);
    }
        
private:
    //! IntersectionIterator on the host grid
    HostLevelIntersectionIterator hostIterator_;
    HostLevelIntersectionIterator endIterator_;

    //! leaf view introduces new non conformity, therefore a host intersection might
    //! consist of several sub-intersections.
    IntersectionListPtr subIntersections_;
    typename IntersectionList::iterator subIntersection_;
};


//! \todo Please doc me !
template<class GridImp>
class SubGridLevelIntersectionIterator :
        private SubGridBasicIntersectionIterator<GridImp>
{
    
    enum {dim=GridImp::dimension};
    
    enum {dimworld=GridImp::dimensionworld};
    
    // The type used to store coordinates
    typedef typename GridImp::ctype ctype;
    
    typedef typename GridImp::HostGridType::template Codim<0>::Entity::LevelIntersectionIterator HostLevelIntersectionIterator;
    
public:

    typedef typename GridImp::template Codim<0>::EntityPointer EntityPointer;
    typedef typename GridImp::template Codim<1>::Geometry Geometry;
    typedef typename GridImp::template Codim<1>::LocalGeometry LocalGeometry;
    typedef typename GridImp::template Codim<0>::Entity Entity;
    typedef Dune::Intersection<const GridImp, Dune::SubGridLevelIntersectionIterator> Intersection;

    SubGridLevelIntersectionIterator(const GridImp* subGrid,
        const HostLevelIntersectionIterator& hostIterator)
        : SubGridBasicIntersectionIterator<GridImp>(subGrid),
          selfLocal_(NULL), neighborLocal_(NULL), intersectionGlobal_(NULL),
          hostIterator_(hostIterator)
          
    {}

    //! equality
    bool equals(const SubGridLevelIntersectionIterator<GridImp>& other) const {
        return hostIterator_ == other.hostIterator_;
    }

    //! prefix increment
    void increment() {
        ++hostIterator_;

        // Delete intersection geometry objects, if present
        if (intersectionGlobal_ != NULL) {
            delete intersectionGlobal_;
            intersectionGlobal_ = NULL;
        }

        if (selfLocal_ != NULL) {
            delete selfLocal_;
            selfLocal_ = NULL;
        }

        if (neighborLocal_ != NULL) {
            delete neighborLocal_;
            neighborLocal_ = NULL;
        }

    }

    //! \brief dereferencing
    const Intersection & dereference() const {
        return reinterpret_cast<const Intersection&>(*this);
    }
        
    //! return EntityPointer to the Entity on the inside of this intersection
    //! (that is the Entity where we started this Iterator)
    EntityPointer inside() const {
        return SubGridEntityPointer<0,GridImp> (this->subGrid_, hostIterator_->inside());
    }
        
    //! return EntityPointer to the Entity on the outside of this intersection
    //! (that is the neighboring Entity)
    EntityPointer outside() const {
        if (!hostIterator_->neighbor()
            || !this->subGrid_->template contains<0>(hostIterator_->outside()))
            DUNE_THROW(GridError,"no neighbor found in outside()");
                
        return SubGridEntityPointer<0,GridImp> (this->subGrid_, hostIterator_->outside());
    }
        
    /** \brief return true if intersection is with boundary.

        Warning: This method uses geometric information and will not work properly in all conceivable
        cases!!  If you find a better way to implement this please let me know.
    */
    bool boundary () const {
//             std::cout << "Checking boundary():   normal: " << unitOuterNormal(FieldVector<double,1>(0)) 
//                       << "   center: " << hostIterator_->inside()->geometry().global(FieldVector<double,2>(0.5))
//                       << std::endl;

        // A boundary intersection of the host grid is also a boundary intersection of the subgrid
        if (hostIterator_->boundary())
            return true;

        // An intersection is boundary if its last ancestor does not have a neighbor,
        HostLevelIntersectionIterator fatherFace = this->getFatherFace(hostIterator_, SubGridBasicIntersectionIterator<GridImp>::coarsestFace);
        // We have found the ancestor face on the coarsest level.  If this is boundary we are boundary
        return !fatherFace->neighbor() || !this->subGrid_->template contains<0>(fatherFace->outside());
    }
        
    //! return true if across the edge an neighbor on this level exists
    bool neighbor () const {
        return hostIterator_->neighbor()
            && this->subGrid_->template contains<0>(hostIterator_->outside());
    }
        
    /** \brief Return information about the Boundary
        \todo This method always returns 1, because it is unclear what it should really do
    */
    int boundaryId () const {
        //DUNE_THROW(NotImplemented, "SubGridLevelIntersectionIterator::boundaryId()");
        return 1;
    }

    /** \brief Get the geometry type */
    GeometryType type () const {
        return hostIterator_->type();
    }
            
    /** \brief Return true if this intersection is conforming */
    bool conforming () const {
        return !neighbor() || hostIterator_->conforming();
    }
              
    //! intersection of codimension 1 of this neighbor with element where
    //! iteration started.
    //! Here returned element is in LOCAL coordinates of the element
    //! where iteration started.
    const LocalGeometry& intersectionSelfLocal () const {
        if (selfLocal_ == NULL)
            selfLocal_ = new SubGridMakeableGeometry<dim-1,dim,GridImp>(hostIterator_->intersectionSelfLocal());
                
        return *selfLocal_;
    }
        
    //! intersection of codimension 1 of this neighbor with element where iteration started.
    //! Here returned element is in LOCAL coordinates of neighbor
    const  LocalGeometry& intersectionNeighborLocal () const {
        if (neighborLocal_ == NULL)
            neighborLocal_ = new SubGridMakeableGeometry<dim-1,dim,GridImp>(hostIterator_->intersectionNeighborLocal());
                
        return *neighborLocal_;
    }
        
    //! intersection of codimension 1 of this neighbor with element where iteration started.
    //! Here returned element is in GLOBAL coordinates of the element where iteration started.
    const Geometry& intersectionGlobal () const {
        if (intersectionGlobal_ == NULL)
            intersectionGlobal_ = new SubGridMakeableGeometry<dim-1,dimworld,GridImp>(hostIterator_->intersectionGlobal());
                
        return *intersectionGlobal_;
    }
        
    //! local number of codim 1 entity in self where intersection is contained in
    int numberInSelf () const {
        return hostIterator_->numberInSelf();
    }
        
    //! local number of codim 1 entity in neighbor where intersection is contained
    int numberInNeighbor () const {
        if (!neighbor())
            DUNE_THROW(GridError, "There is no neighbor!");

        return hostIterator_->numberInNeighbor();
    }
    
    //! return outer normal, this should be dependent on local
    //! coordinates for higher order boundary
    const FieldVector<ctype, dimworld> outerNormal (const FieldVector<ctype, dim-1>& local) const {
        return hostIterator_->outerNormal(local);
    }

    //! return outer normal, this should be dependent on local
    //! coordinates for higher order boundary
    const FieldVector<ctype, dimworld> integrationOuterNormal (const FieldVector<ctype, dim-1>& local) const {
        return  hostIterator_->integrationOuterNormal(local);
    }

    //! return outer normal, this should be dependent on local
    //! coordinates for higher order boundary
    const FieldVector<ctype, dimworld> unitOuterNormal (const FieldVector<ctype, dim-1>& local) const {
        return hostIterator_->unitOuterNormal(local);
    }
        
private:
    
    //! pointer to element holding the selfLocal and selfGlobal information.
    //! This element is created on demand.
    mutable SubGridMakeableGeometry<dim-1,dim,GridImp>* selfLocal_;
    mutable SubGridMakeableGeometry<dim-1,dim,GridImp>* neighborLocal_;
    
    //! pointer to element holding the neighbor_global and neighbor_local
    //! information.
    mutable SubGridMakeableGeometry<dim-1,dimworld,GridImp>* intersectionGlobal_;
    
    HostLevelIntersectionIterator hostIterator_;

};


}  // namespace Dune

#endif
