#ifndef DUNE_SUBGRIDTOOLS_HH
#define DUNE_SUBGRIDTOOLS_HH

/** \file
* \brief SubGrid tool classes for interaction with the hostgrid
*/

#include <vector>

#include <dune/disc/shapefunctions/lagrangeshapefunctions.hh>

namespace Dune {




//! \todo Please doc me !
template<class GridType, class Vector>
class SubGridP1Interpolator
{
    public:
        typedef typename GridType::template Codim<0>::EntityPointer ElementPointer;
        typedef typename GridType::HostGridType HostGridType;
        typedef typename HostGridType::template Codim<0>::EntityPointer HostElementPointer;
        
        enum {dim = GridType::dimension};
        
        
        /** \brief Interpolate subgrid P1 function to hostgrid
        *
        * This is a shortcut for creating an (temporary) interpolator object and
        * handing it to the transfer method in the subgrid
        */
        static void interpolate(const GridType& _grid, const Vector& _subGridFunction, Vector& _hostGridFunction)
        {
            typedef SubGridP1Interpolator<GridType, Vector> MyType;
            
            MyType interpolator(_grid, _subGridFunction, _hostGridFunction);
            _grid.template transfer<MyType>(interpolator);
        }
        
        
        //! \todo Please doc me !
        SubGridP1Interpolator(const GridType& _grid, const Vector& _subGridFunction, Vector& _hostGridFunction) :
            grid(_grid),
            hostgrid(_grid.getHostGrid()),
            subGridFunction(_subGridFunction),
            hostGridFunction(_hostGridFunction)
        {}
        
        
        //! Prepare hostgrid vector
        void pre()
        {
            processed.resize(hostgrid.size(dim), false);
            hostGridFunction.resize(hostgrid.size(dim));
            hostGridFunction = 0.0;
        }
        
        
        //! \todo Please doc me !
        void transfer(const ElementPointer& element, const HostElementPointer& hostElement)
        {
            const Dune::LagrangeShapeFunctionSet<double, double, dim>& subGridBaseSet
                = Dune::LagrangeShapeFunctions<double, double, dim>::general(element->type(), 1);
            const int subGridBaseSetSize = subGridBaseSet.size();
            
            for (int i=0; i<hostElement->template count<dim>(); ++i)
            {
                int hgIndex = hostgrid.leafIndexSet().template subIndex<dim>(*hostElement, i);
                if (not(processed[hgIndex]))
                {
                    FieldVector<double, dim> local = element->geometry().local(hostElement->geometry()[i]);
                    for (int j=0; j<subGridBaseSetSize; ++j)
                    {
                        int sgIndex = grid.leafIndexSet().template subIndex<dim>(*element, j);
                        typename Vector::block_type interpolatedValue = subGridFunction[sgIndex];
                        interpolatedValue *= subGridBaseSet[j].evaluateFunction(0, local);
                        hostGridFunction[hgIndex] += interpolatedValue;
                    }
                    processed[hgIndex] = true;
                }
            }
            return;
        }
        
        
        //! does nothing
        void post()
        {}
        
        
    private:
        const GridType& grid;
        const HostGridType& hostgrid;
        
        //! \todo Please doc me !
        std::vector<bool> processed;
        
        //! \todo Please doc me !
        const Vector& subGridFunction;
        
        //! \todo Please doc me !
        Vector& hostGridFunction;
};




//! \todo Please doc me !
template<class GridType, class Vector>
class SubGridP1Restrictor
{
    public:
        typedef typename GridType::template Codim<0>::EntityPointer ElementPointer;
        typedef typename GridType::HostGridType HostGridType;
        typedef typename HostGridType::template Codim<0>::EntityPointer HostElementPointer;
        
        enum {dim = GridType::dimension};
        
        
        /** \brief Restrict hostgrid P1 functional to subgrid
        *
        * This is a shortcut for creating an (temporary) restrictor object and
        * handing it to the transfer method in the subgrid
        */
        static void restrict(const GridType& _grid, const Vector& _hostGridFunction, Vector& _subGridFunction)
        {
            typedef SubGridP1Restrictor<GridType, Vector> MyType;
            
            MyType restrictor(_grid, _hostGridFunction, _subGridFunction);
            _grid.template transfer<MyType>(restrictor);
        }
        
        
        //! \todo Please doc me !
        SubGridP1Restrictor(const GridType& _grid, const Vector& _hostGridFunction, Vector& _subGridFunction) :
            grid(_grid),
            hostgrid(_grid.getHostGrid()),
            hostGridFunction(_hostGridFunction),
            subGridFunction(_subGridFunction)
        {}
        
        
        //! Prepare subgrid vector
        void pre()
        {
            processed.resize(hostgrid.size(dim), false);
            subGridFunction.resize(grid.size(dim));
            subGridFunction = 0.0;
        }

        
        //! \todo Please doc me !
        void transfer(const ElementPointer& element, const HostElementPointer& hostElement)
        {
            const Dune::LagrangeShapeFunctionSet<double, double, dim>& subGridBaseSet
                = Dune::LagrangeShapeFunctions<double, double, dim>::general(element->type(), 1);
            const int subGridBaseSetSize = subGridBaseSet.size();
            
            for (int i=0; i<hostElement->template count<dim>(); ++i)
            {
                int hgIndex = hostgrid.leafIndexSet().template subIndex<dim>(*hostElement, i);
                if (not(processed[hgIndex]))
                {
                    FieldVector<double, dim> local = element->geometry().local(hostElement->geometry()[i]);
                    for (int j=0; j<subGridBaseSetSize; ++j)
                    {
                        int sgIndex = grid.leafIndexSet().template subIndex<dim>(*element, j);
                        subGridFunction[sgIndex] += subGridBaseSet[j].evaluateFunction(0, local) * hostGridFunction[hgIndex];
                    }
                    processed[hgIndex] = true;
                }
            }
            return;
        }
        
        
        //! does nothing
        void post()
        {}
        
        
    private:
    
        const GridType& grid;
        const HostGridType& hostgrid;
        
        //! \todo Please doc me !
        std::vector<bool> processed;
        
        //! \todo Please doc me !
        const Vector& hostGridFunction;
        
        //! \todo Please doc me !
        Vector& subGridFunction;
};


} // namespace Dune

#endif
